// 引入 TreeRenderResult 类型，该类型定义在 @/lib/XMLProcessor/src/typing 模块中
import { NewTreeModification, OldTreeModification, TreeRenderResult, TreeRenderResultFlatted } from '@/lib/XMLProcessor/src/typing'
import { md5 } from 'js-md5'
import { getUUID } from '../utils/uuid.ts'
import { dualCompare } from '@/lib/XMLProcessor/src/core/dualCompare.ts'
import {cloneDeep} from 'lodash'

// 定义 Processing 类，用于处理 XML 数据
export class Processing {
    // 私有属性，用于解析 XML 字符串为 DOM 对象
    private domParser: DOMParser
    private xmlSerializer: XMLSerializer

    // 构造函数，初始化 DOMParser 实例
    constructor() {
        this.domParser = new DOMParser()
        this.xmlSerializer = new XMLSerializer()
    }

    serializeXML(xmlDOM: Document): string {
        return this.xmlSerializer.serializeToString(xmlDOM)
    }

    private flattenTree(renderResult: TreeRenderResult[]): TreeRenderResultFlatted[] {
        const resp: TreeRenderResultFlatted[] = []
        for (const item of renderResult) {
            resp.push({
                key: item.key,
                label: item.label,
                hash: item.hash,
                index: item.index,
                chained: item.chained,
                textContent: item.textContent,
                node: item.node
            })
            if (item.children) {
                resp.push(...this.flattenTree(item.children))
            }
        }
        return resp
    }

    private buildTreeFromFlatted(
        flatted: TreeRenderResultFlatted[],
    ): TreeRenderResult[] {
        const result: TreeRenderResult[] = []
        if (!flatted.length) {
            return result
        }
        const flattedTreeSorted = flatted.toSorted((a, b) => {
            return a.chained.length - b.chained.length;
        });
        const buildTree = (singleItem: TreeRenderResultFlatted, flattedDataSet: TreeRenderResultFlatted[]) => {
            const singleItemRendered: TreeRenderResult = Object.assign({}, singleItem, {children: []});
            const currentChainedLength = singleItemRendered.chained.length;
            const maxChainedLength = Math.max(...flattedDataSet.map(a=>a.chained.length));
            const currentChainedPrefix = singleItemRendered.chained.join('.');
            for (let i = currentChainedLength + 1; i<= maxChainedLength; i++) {
                const allSubItems = flattedDataSet.filter(a=>a.chained.length === i && a.chained.join('.').startsWith(currentChainedPrefix));
                if (allSubItems.length) {
                    for (let j = flattedDataSet.length - 1; j>=0; j--) {
                        if (allSubItems.includes(flattedDataSet[j])) {
                            flattedDataSet.splice(j, 1)
                        }
                    }
                    for (const item of allSubItems) {
                        const subTree = buildTree(item, flattedDataSet);
                        singleItemRendered.children?.push(subTree)
                    }
                }
            }
            return singleItemRendered
        }
        for (const item of flattedTreeSorted) {
            result.push(buildTree(item, flattedTreeSorted))
        }
        return result
    }

    dualCompareFromString(
        xmlStringOld: string,
        xmlStringNew: string,
        bothHandledNode: string[],
        contentHoldNode: string[]
    ): {
        dataForNew: NewTreeModification
        dataForOld: OldTreeModification
        domOld: Document
        domNew: Document
        treeOld: TreeRenderResult[]
        treeNew: TreeRenderResult[]
        xmlContentOld: string
        xmlContentNew: string
    } {
        const treeA = this.domParser.parseFromString(xmlStringOld, 'text/xml')
        const treeB = this.domParser.parseFromString(xmlStringNew, 'text/xml')
        const resultA = this.innerHandle(treeA.documentElement, bothHandledNode, contentHoldNode)
        const resultB = this.innerHandle(treeB.documentElement, bothHandledNode, contentHoldNode)
        treeA.documentElement.querySelectorAll(':not([data-key])').forEach((node) => node.remove())
        treeB.documentElement.querySelectorAll(':not([data-key])').forEach((node) => node.remove())
        // 扁平化所有节点
        const resultAFlatted = this.flattenTree(resultA)
        const resultBFlatted = this.flattenTree(resultB)

        const results = dualCompare(resultAFlatted, resultBFlatted).filter((a) => a.type !== 'same')

        console.log('after-compare', results)

        // 分组构建树
        const allDeletedTree = results.filter((item) => item.type === 'deleted').map(a=>a.item);

        const allAddedTree = results.filter((item) => item.type === 'added').map(a=>a.item);

        const dataForOld: OldTreeModification = {
            Deleted: this.buildTreeFromFlatted(allDeletedTree)
        }
        const dataForNew: NewTreeModification = {
            Added: this.buildTreeFromFlatted(allAddedTree)
        }

        return {
            dataForNew,
            dataForOld,
            domOld: treeA,
            domNew: treeB,
            treeOld: resultA,
            treeNew: resultB,
            xmlContentOld: this.serializeXML(treeA),
            xmlContentNew: this.serializeXML(treeB)
        }
    }

    // 处理 XML 字符串的方法，返回树形数据和 XML DOM 对象
    processXML(
        xmlContent: string,
        handledNode: string[],
        contentHoldNode: string[]
    ): {
        treeData: TreeRenderResult[]
        xmlDOM: Document
        xmlContent: string
    } {
        // 使用 DOMParser 解析 XML 字符串
        const parsedTree = this.domParser.parseFromString(xmlContent, 'text/xml')
        // 获取 XML 文档的根元素
        const rootElement = parsedTree.documentElement
        const result = {
            // 调用 innerHandle 方法处理根元素，生成树形数据
            treeData: this.innerHandle(rootElement, handledNode, contentHoldNode),
            // 返回解析后的 XML DOM 对象
            xmlDOM: parsedTree
        }
        parsedTree.documentElement.querySelectorAll(':not([data-key])').forEach((node) => node.remove())
        return {
            ...result,
            xmlContent: this.serializeXML(parsedTree)
        }
    }

    // 异步处理 XML 文件的方法，返回包含树形数据和 XML DOM 对象的 Promise
    async processFile(
        xmlFile: File,
        handledNode: string[],
        contentHoldNode: string[]
    ): Promise<{
        treeData: TreeRenderResult[]
        xmlDOM: Document
        xmlContent: string
    }> {
        // 检查文件是否为有效的 XML 文件
        if (!(xmlFile instanceof File) || !xmlFile.type.includes('xml')) {
            return Promise.reject('The file is not a valid XML file.')
        }
        return new Promise((resolve, reject) => {
            const fr = new FileReader()
            // 当文件读取完成时触发的事件处理函数
            fr.onload = () => {
                const content = fr.result as string
                try {
                    // 调用 processXML 方法处理读取的文件内容
                    const { treeData, xmlDOM, xmlContent } = this.processXML(content, handledNode, contentHoldNode)

                    // 解析成功，返回树形数据和 XML DOM 对象
                    resolve({ treeData, xmlDOM, xmlContent })
                } catch (error) {
                    // 解析失败，拒绝 Promise 并返回错误信息
                    reject(error)
                }
            }
            // 当文件读取出错时触发的事件处理函数
            fr.onerror = () => {
                reject('Error reading file.')
            }
            // 以文本形式读取文件
            fr.readAsText(xmlFile)
        })
    }

    // 私有方法，用于递归处理单个 DOM 节点，生成树形数据
    private innerHandle(domNode: Element, handledNode: string[], contentHoldNode: string[]): TreeRenderResult[] {
        let treeData: TreeRenderResult[] = []
        // 生成唯一的 key
        let targetKey = domNode.getAttribute('data-key')
        if (!targetKey) {
            targetKey = getUUID()
        }
        // 创建树形数据项
        const treeItem: TreeRenderResult = {
            key: targetKey,
            label: domNode.nodeName,
            children: [],
            index: 0,
            chained: [0],
            hash: contentHoldNode.includes(domNode.nodeName) ? md5.hex(domNode.textContent! || '') : md5(domNode.nodeName),
            textContent: domNode.textContent || '',
            node: domNode
        }
        // 为 DOM 节点设置 data-key 属性
        domNode.setAttribute('data-key', targetKey)
        domNode.setAttribute('data-w-e-type', domNode.nodeName)
        domNode.setAttribute('data-indent-level', '0')
        domNode.setAttribute('data-chained', treeItem.chained.join('$'))
        domNode.setAttribute('data-hash', treeItem.hash)
        // 递归处理子节点
        treeItem.children = this.innerGroupHandle(domNode.children, handledNode, [0], contentHoldNode)
        // 将树形数据项添加到结果数组中
        treeData.push(treeItem)

        return treeData
    }

    // 私有方法，用于递归处理一组 DOM 节点，生成树形数据
    private innerGroupHandle(nodeSet: HTMLCollection, handledNode: string[], startChained: number[], contentHoldNode: string[]): TreeRenderResult[] {
        let targetResult: TreeRenderResult[] = []
        let realIndex = -1
        // 遍历节点集合
        for (let i = 0; i < nodeSet.length; i++) {
            const node = nodeSet[i] as Element
            // 检查节点是否在处理列表中
            if (!handledNode.includes(node.nodeName)) {
                // 如果节点有子节点，递归处理子节点
                if (node.children.length) {
                    targetResult.push(...this.innerGroupHandle(node.children, handledNode, startChained, contentHoldNode))
                }
                continue
            }
            realIndex++
            // 生成唯一的 key
            let targetKey = node.getAttribute('data-key')
            if (!targetKey) {
                targetKey = getUUID()
            }
            // 创建树形数据项
            const treeItem: TreeRenderResult = {
                key: targetKey,
                label: node.nodeName,
                index: realIndex,
                chained: [...startChained, realIndex],
                hash: contentHoldNode.includes(node.nodeName) ? md5.hex(node.textContent! || '') : md5(node.nodeName),
                textContent: node.textContent || '',
                node: node
            }
            // 为 DOM 节点设置 data-key 属性
            node.setAttribute('data-key', targetKey)
            node.setAttribute('data-w-e-type', node.nodeName)
            node.setAttribute('data-indent-level', startChained.length.toString())
            node.setAttribute('data-chained', treeItem.chained.join('$'))
            node.setAttribute('data-hash', treeItem.hash)
            // 如果节点有子节点，递归处理子节点
            if (node.children.length > 0) {
                treeItem.children = this.innerGroupHandle(node.children, handledNode, [...startChained, realIndex], contentHoldNode)
            }
            // 将树形数据项添加到结果数组中
            targetResult.push(treeItem)
        }
        return targetResult
    }
}
